#!/usr/bin/python3
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

"""
Used to enhance profiles. Currently it adds regression/improvement
markers to the profiles.
"""

import copy
import json
import pathlib


class ProfileEnhancer:
    def __init__(self, output_dir):
        self.output_dir = pathlib.Path(output_dir).resolve()

    def get_geckomain_thread(self, data_after):
        """Searches through a profile to find a GeckoMain thread.

        At the same time, it also finds a start time that is later
        used for correcting the delta that gets added to the visual
        metrics regression/improvement markers. See here:
        https://github.com/firefox-devtools/profiler/blob/
        a69c98867d185c069560c5b1aa56da8c61ac97b9/src/profile-logic/
        process-profile.js#L1398-L1399
        """
        gecko_main = None
        start_time = None

        for proc in data_after["processes"]:
            for thread in proc["threads"]:
                if thread["name"] == "GeckoMain":
                    gecko_main = thread
                    start_time = proc["meta"]["startTime"]
                    break

        return gecko_main, start_time

    def get_test_category(self, data_after):
        """Finds, and returns the index of the Test category."""
        for i, cat in enumerate(data_after["meta"]["categories"]):
            if cat["name"].lower() == "test":
                return i
        raise Exception("Could not find the test category in the given profile.")

    def insert_into_string_table(self, gecko_main):
        gecko_main["stringTable"].append("Performance Changes Detected (from ProfileEnhancer)")
        return len(gecko_main["stringTable"]) - 1

    def detect_changes(self, vismet_before_ns, vismet_after_ns, threshold=5):
        """Returns the changes (improvements, AND regressions) found.

        We go through both lists and once we find a mismatch, we begin
        marking either a regression, OR, an improvment and then store it
        once we hit an equalization point.

        DO NOT use approximations here because the visual completeness
        can absolutely change in flash. This means that we begin counting
        a regression, then stop when we hit a point that is unregressed
        in the other.

        Use the threshold to determine how large the distance in progress
        must be to be classified as a change.
        """ 
        changes = []

        # Standardize the timestamps
        vismet_before = []
        vismet_after = []
        for prog in vismet_before_ns:
            new_prog = copy.deepcopy(prog)
            new_prog["timestamp"] = prog["timestamp"] - vismet_before_ns[0]["timestamp"]
            vismet_before.append(new_prog)
        for prog in vismet_after_ns:
            new_prog = copy.deepcopy(prog)
            new_prog["timestamp"] = prog["timestamp"] - vismet_after_ns[0]["timestamp"]
            vismet_after.append(new_prog)

        start_ts = None
        end_ts = None
        curr_ind = 0
        for i, prog in enumerate(vismet_before):
            if prog["percent"] == vismet_after[curr_ind]["percent"]:
                curr_ind += 1
            elif (
                prog["percent"] > vismet_after[curr_ind]["percent"] and
                prog["timestamp"] <= vismet_after[curr_ind]["timestamp"]
            ):
                # Create a regression
                # From this point until we find an inflection in the progress
                # we no longer care about the timestamp. This is because we already
                # know we're slower in the vismet_after starting from now, and
                # until the progress flips again. The same is true for an improvement.
                start_ts = vismet_after[curr_ind]["timestamp"]
                while (
                    prog["percent"] > vismet_after[curr_ind]["percent"] and
                    curr_ind + 1 < len(vismet_after)
                ):
                    curr_ind += 1
                    end_ts = vismet_after[curr_ind]["timestamp"]

                changes.append({
                    "type": "regression",
                    "start": start_ts + vismet_after_ns[0]["timestamp"],
                    "end": end_ts + vismet_after_ns[0]["timestamp"],
                })
            elif (
                prog["percent"] < vismet_after[curr_ind]["percent"] and
                prog["timestamp"] >= vismet_after[curr_ind]["timestamp"]
            ):
                # Create an improvement
                start_ts = vismet_after[curr_ind]["timestamp"]
                while (
                    prog["percent"] < vismet_after[curr_ind]["percent"] and 
                    curr_ind + 1 < len(vismet_after)
                ):
                    curr_ind += 1
                    end_ts = vismet_after[curr_ind]["timestamp"]

                changes.append({
                    "type": "improvement",
                    "start": start_ts + vismet_after_ns[0]["timestamp"],
                    "end": end_ts + vismet_after_ns[0]["timestamp"],
                })

            if curr_ind >= len(vismet_after):
                # Finished searching
                break

        # Merge the changes we've found in case there's overlap,
        # this could be resolved above but I prefer getting steps
        # of regressions as an intermediate value for debugging
        merged_changes = []
        curr_change = None
        for change in changes:
            if curr_change is None:
                curr_change = change
            elif (
                curr_change["type"] == change["type"] and 
                curr_change["end"] == change["start"]
            ):
                curr_change["end"] = change["end"]
            else:
                merged_changes.append(curr_change)
                curr_change = change
        if curr_change:
            merged_changes.append(curr_change)

        return merged_changes

    def run(self, profile_before, profile_after):
        """
        This method will go through the visual metrics in both
        profiles to try to find all areas where there is either
        a regression or improvement and mark them into a new
        profile in the given output directory.
        """
        profile_before = pathlib.Path(profile_before).resolve()
        with profile_before.open() as f:
            data_before = json.load(f)

        profile_after = pathlib.Path(profile_after).resolve()
        with profile_after.open() as f:
            data_after = json.load(f)

        gecko_main, start_time = self.get_geckomain_thread(data_after)

        # Calculate regression ranges
        vismets_before = data_before["meta"].get("visualMetrics", None)
        vismets_after = data_after["meta"].get("visualMetrics", None)

        # Calculate a regression/improvment range for each visual progress metric
        psi_changes = self.detect_changes(
            vismets_before.get("PerceptualSpeedIndexProgress", []),
            vismets_after.get("PerceptualSpeedIndexProgress", [])
        )
        print(psi_changes)

        csi_changes = self.detect_changes(
            vismets_before.get("ContentfulSpeedIndexProgress", []),
            vismets_after.get("ContentfulSpeedIndexProgress", [])
        )
        print(csi_changes)

        si_changes = self.detect_changes(
            vismets_before.get("VisualProgress", []),
            vismets_after.get("VisualProgress", [])
        )
        print(si_changes)

        # Insert the change markers into the after profile
        perfchange_name_ind = self.insert_into_string_table(gecko_main)
        test_cat_ind = self.get_test_category(data_after)
        diff = start_time - data_after["meta"]["startTime"]
        for prog_type, changes in [
            ("PerceptualSpeedIndex", psi_changes),
            ("ContenfulSpeedIndex", csi_changes),
            ("SpeedIndex", si_changes)
        ]:
            for change in changes:
                print(f"{prog_type}: {change}")
                gecko_main["markers"]["data"].append(
                    [
                        perfchange_name_ind,
                        change["start"] - diff,
                        change["end"] - diff,
                        1,
                        test_cat_ind,
                        {
                            "type": "Task",
                            "name": f"{prog_type} - {change['type']}",
                        }
                    ]
                )

        # Save the data
        profile_after_name = profile_after.name.split(".")[0]
        res = pathlib.Path(self.output_dir, f"{profile_after_name}-enhanced.json.gz")

        print(f"Saving enhanced profile to {str(res.resolve())}")
        with res.open("w") as f:
            json.dump(data_after, f)



#!/usr/bin/python3
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

"""
Used to detect regressions in-tree.
"""

import copy
import json
import numpy as np
import pathlib
import scipy
import shutil

from moz_measure_noise import step_detector

from mozperftest_tools.side_by_side import SideBySide
from mozperftest_tools.utils.artifact_downloader import artifact_downloader
from mozperftest_tools.utils.task_processor import (
    get_task_data_paths,
    match_vismets_with_videos,
    sorted_nicely,
)
from mozperftest_tools.utils.utils import (
    finish_same_line,
    find_task_group_id,
    get_pushes,
    get_revision_json,
    write_same_line,
)

"""
TODO:
* Add sliding window parameter to new method.
* Detect across all tests, and platforms.

"""

ALLOWED_METHODS = ["mwu"]
MAX_WINDOW = 1


class MethodNotFoundError(Exception):
    """Raised when a specified detection method is not found."""

    pass


class ZeroDepthError(Exception):
    """Raised when the auto-computed depth is zero."""

    pass


class BranchMismatchError(Exception):
    """Raised when branches do not match."""

    pass


class NoDataError(Exception):
    """Raised when we can't find enough data."""

    pass


class RegressionDetector(SideBySide):
    def __init__(self, output_dir, method="mwu"):
        self._output_dir = pathlib.Path(output_dir).resolve()
        self._cache = None

        self.method = method
        if method not in ALLOWED_METHODS:
            raise MethodNotFoundError(f"Unknown method: {method}")

    def _gather_revisions(self, start, end, branch, depth):
        """Returns all revisions/pushes found in a given range.

        Either the entire range of pushes from start (base) to end (new)
        is returned or a maximum of "depth" number of pushes (end-depth).
        """
        end_id = get_revision_json(end, branch=branch, cache=self._cache)["pushid"]
        if depth == -1:
            # We need to determine the depth to search
            if start == end:
                raise ZeroDepthError(
                    "The starting, and ending revisions can't be the same when"
                    "depth is set to -1 (auto-computed)."
                )
            start_info = get_revision_json(start, branch=branch, cache=self._cache)
            depth = (int(end_id) - int(start_info["pushid"])) + 1
            print(f"Using an auto-computed depth of {depth} to gather revisions")
        pushes = get_pushes(
            branch,
            end_id,
            depth,
            True,
            cache=self._cache,
        )

        push_revisions = []
        for push_id in sorted_nicely(list(pushes.keys())):
            push_revisions.append(pushes[push_id]["changesets"][-1])

        return push_revisions

    def _open_and_organize_perfherder(self, files):
        def _open_perfherder(filen):
            with open(filen) as f:
                return json.load(f)

        res = {"cold": {}, "warm": {}}

        for filen in files:
            data = _open_perfherder(filen)

            for suite in data["suites"]:
                pl_type = "warm"
                if "cold" in suite["extraOptions"]:
                    pl_type = "cold"

                for subtest in suite["subtests"]:
                    if "cputime" in subtest["name"].lower():
                        continue
                    # Each entry here will be a single retrigger of
                    # the test for the requested metric (ordered
                    # based on the `files` ordering)
                    res[pl_type].setdefault(subtest["name"], []).extend(
                        subtest["replicates"]
                    )

        return res

    def detect_all_changes(
        self,
        base_branch="autoland",
        new_branch="autoland",
        base_revision="",
        new_revision="",
        depth=None,
        search_crons=True,
        overwrite=True,
        skip_download=False,
    ):
        """Detect all changes in all tests in the given revisions."""
        pass

    def detect_changes(
        self,
        test_name="",
        new_test_name="",
        platform="",
        new_platform="",
        base_branch="autoland",
        new_branch="autoland",
        base_revision="",
        new_revision="",
        depth=None,
        search_crons=True,
        overwrite=True,
        skip_download=False,
    ):
        """
        This method is the main method of the class and uses all other methods to make the actual side-by-side comparison.

        :param test_name: The name of the test task to get videos from.
        :param new_test_name: The name of the test task to get videos from in the new revision.
        :param platform: Platform to return results for.
        :param new_platform: Platform to return results for in the new revision.
        :param vismetPath: Paths to visualmetrics.py for step chart generation.
        :param metric: Metric to use for side-by-side comparison.
        :param base_branch: Branch to search for the base revision.
        :param new_branch: Branch to search for the new revision.
        :param base_revision: The base revision to compare a new revision to
        :param new_revision: The base revision to compare a new revision to.
        :param search_crons: If set, we will search for the tasks within the cron jobs as well.
        :param cold: If set, we'll only look at cold pageload tests.
        :param warm: If set, we'll only look at warm pageload tests.
        :param most_similar: If set, we'll search for a video pairing that is the most similar.
        :param overwrite: If set, the downloaded task group data will be deleted before it gets re-downloaded.
        :param skip_download: If set, we won't try to download artifacts again and we'll try using what already exists in the output folder.
        :param original: If set, use the original videos in the side-by-side instead of the postprocessed videos.
        :return:
        """
        # Parse the given output argument
        output = pathlib.Path(self._output_dir)
        if not output.suffixes:
            output.mkdir(parents=True, exist_ok=True)
        else:
            filename = output.name
            output = output.parents[0]
            output.mkdir(parents=True, exist_ok=True)

        # Make a cache for the push revision downloads which can require
        # a lot of artifacts to find enough data
        self._cache = pathlib.Path(output, ".cache")
        self._cache.mkdir(exist_ok=True)

        revisions = [
            (base_revision, base_branch, test_name, platform),
            (new_revision, new_branch, new_test_name, new_platform),
        ]
        if depth is not None:
            if base_branch != new_branch:
                raise BranchMismatchError(
                    "Can't compare using depth across multiple branches! "
                    f"{base_branch} != {new_branch}"
                )
            push_revisions = self._gather_revisions(
                base_revision, new_revision, base_branch, depth
            )

            revisions = list(
                zip(
                    push_revisions,
                    [base_branch] * len(push_revisions),
                    [test_name] * len(push_revisions),
                    [platform] * len(push_revisions),
                )
            )

        return self.compare_revisions(revisions, output, overwrite, skip_download)

    def compare_revisions(self, revisions, output, overwrite, skip_download):
        """Compare multiple revisions together.

        Revisions must be ordered with the oldest first (base),
        and newest last (new).
        """

        # Get the task group IDs for the revisions
        all_revision_ids = []
        for (revision, branch, _, _) in revisions:
            all_revision_ids.append(
                find_task_group_id(
                    revision, branch, search_crons=True, cache=self._cache
                )
            )

        all_revision_dirs = [
            [pathlib.Path(output, revid) for revid in revision_ids]
            for revision_ids in all_revision_ids
        ]

        if overwrite:
            for task_dirs in all_revision_dirs:
                for task_dir in task_dirs:
                    if task_dir.exists():
                        print("Removing existing task group folder: %s" % str(task_dir))
                        shutil.rmtree(str(task_dir))

        # Download the artifacts
        new_revisions = []
        all_revision_data_paths = []
        for i, (revision, _, test_name, platform) in enumerate(revisions):
            revision_ids = all_revision_ids[i]
            if skip_download and not overwrite:
                try:
                    data_paths = self._search_for_paths(revision_ids, "perfherder-data")
                except (AttributeError, FileNotFoundError) as e:
                    print("No data was found for revision %s" % revision)
                    data_paths = []
            else:
                data_paths = []
                for revision_id in revision_ids:
                    if data_paths:
                        break
                    artifact_downloader(
                        revision_id,
                        output_dir=str(output),
                        test_suites=[test_name],
                        platform=platform,
                        artifact_to_get=["perfherder-data"],
                        unzip_artifact=False,
                        download_failures=False,
                        ingest_continue=False,
                    )
                    try:
                        data_paths = self._search_for_paths(
                            [revision_id], "perfherder-data"
                        )
                    except (AttributeError, FileNotFoundError) as e:
                        print("No data was found for revision %s" % revision)
                        data_paths = []
            if data_paths:
                new_revisions.append(revision)
                all_revision_data_paths.append(data_paths)

        if len([paths for paths in all_revision_data_paths if paths]) < 2:
            print("All paths found: %s" % all_revision_data_paths)
            raise NoDataError("Not enough artifacts downloaded, can't compare! ")

        all_revision_data = []
        for revision_data_paths in all_revision_data_paths:
            all_revision_data.append(
                self._open_and_organize_perfherder(revision_data_paths)
            )

        # We have all the data downloaded, now organize it into lists or metrics
        # and detect any changes.
        all_metric_names = {"warm": {}, "cold": {}}
        for data in all_revision_data:
            for pl_type in ("warm", "cold"):
                for metric in data[pl_type]:
                    all_metric_names[pl_type].setdefault(metric, 0)
                    all_metric_names[pl_type][metric] += 1

        use_step_detector = False
        data_org_by_metric = {"warm": {}, "cold": {}}
        revisions_org_by_metric = {"warm": {}, "cold": {}}
        for i, data in enumerate(all_revision_data):
            for pl_type in ("warm", "cold"):
                for metric in data[pl_type]:
                    if all_metric_names[pl_type][metric] != len(all_revision_data):
                        # This metric doesn't exist in one, or more revisions but
                        # other metrics do exist
                        print(
                            "Missing data for metric %s, found %s, expected %s"
                            % (
                                metric,
                                all_metric_names[pl_type][metric],
                                all_revision_data,
                            )
                        )
                        continue

                    if use_step_detector:
                        data_org_by_metric[pl_type].setdefault(metric, []).extend(
                            data[pl_type][metric]
                        )
                        revisions_org_by_metric[pl_type].setdefault(metric, []).extend(
                            [new_revisions[i]] * len(data[pl_type][metric])
                        )
                    else:
                        data_org_by_metric[pl_type].setdefault(metric, []).append(
                            data[pl_type][metric]
                        )
                        revisions_org_by_metric[pl_type].setdefault(metric, []).append(
                            new_revisions[i]
                        )

        results = {"warm": {}, "cold": {}}
        for pl_type in ("warm", "cold"):
            for metric in data_org_by_metric[pl_type]:
                if use_step_detector:
                    print("Using sliding window method")
                    segments, diffs = step_detector.find_segments(
                        data_org_by_metric[pl_type][metric], 1, 0.03
                    )
                else:
                    print("Using non-sliding window method")
                    segments = [0]
                    diffs = [0]
                    prev_data_group = data_org_by_metric[pl_type][metric][0]
                    for i, data_group in enumerate(
                        data_org_by_metric[pl_type][metric][1:], 1
                    ):
                        m_score = scipy.stats.mannwhitneyu(
                            prev_data_group,
                            data_group,
                            alternative="two-sided",
                        )
                        print("\nCurrent score: ", m_score)
                        print("Metric: ", pl_type, metric)
                        print("Revision: ", revisions_org_by_metric[pl_type][metric][i])

                        window = 0
                        while (
                            (window + 1) <= MAX_WINDOW
                            and m_score.pvalue < 0.06
                            and m_score.pvalue > 0.001
                        ):
                            # For scores that are borderline, try to improve it by adding data to both sides.
                            # At the moment, this is very experimental, but it theoretically can improve, and
                            # has improved results already.
                            window += 1
                            print("Recomputing %s" % str(m_score))
                            if i - (window + 1) >= 0:
                                prev_data_group.extend(
                                    data_org_by_metric[pl_type][metric][
                                        i - (window + 1)
                                    ]
                                )
                            if i + window < len(data_org_by_metric[pl_type][metric]):
                                data_group.extend(
                                    data_org_by_metric[pl_type][metric][i + window]
                                )

                            m_score = scipy.stats.mannwhitneyu(
                                prev_data_group,
                                data_group,
                                alternative="two-sided",
                            )
                            print("Recomputed to %s" % str(m_score))
                        print("New score: ", m_score)

                        if m_score.pvalue <= 0.001:
                            # Check if the differences in the median cross the threshold
                            prev_med = np.median(prev_data_group)
                            prev_std = np.std(prev_data_group)
                            curr_med = np.median(data_group)

                            # MWU results is U1, use effect size from here:
                            # https://en.wikipedia.org/wiki/Mann%E2%80%93Whitney_U_test#Rank-biserial_correlation
                            effect_size = (2 * m_score.statistic) / (
                                len(prev_data_group) * len(data_group)
                            )

                            # Get a threshold to apply based on the noise, and
                            # then check if against the %-difference between
                            # the previous and current median. Medians are used here
                            # since MWU is based on them as well.
                            fuzz = 0.02
                            threshold = (prev_std + (prev_med * fuzz)) / prev_med
                            diff = abs(prev_med - curr_med) / prev_med
                            if diff > threshold:
                                print(
                                    f"**** Found difference:",
                                    threshold,
                                    diff,
                                    effect_size,
                                    "*******",
                                )
                                segments.append(i)
                                diffs.append(diff)

                        prev_data_group = data_group

                    # Append the last index for compatibility
                    # with the sliding window method
                    if segments[-1] != i:
                        segments.append(i)

                results[pl_type][metric] = segments, diffs

        all_regressing_revisions = set()
        regressed_metric_revisions = {"warm": {}, "cold": {}}
        for pl_type in ("warm", "cold"):
            for metric in results[pl_type]:
                segments, diffs = results[pl_type][metric]
                if len(segments) <= 2:
                    # There was no regression/improvment
                    continue

                for regressing_index in segments[1:-1]:
                    revision = revisions_org_by_metric[pl_type][metric][
                        regressing_index
                    ]
                    regressed_metric_revisions[pl_type].setdefault(
                        metric, {}
                    ).setdefault(revision, []).append(regressing_index)
                    all_regressing_revisions |= set([revision])

        print("Regressed, and ordered revisions:")
        for pl_type in ("warm", "cold"):
            for metric in results[pl_type]:
                print(f"{pl_type} {metric}")
                if metric not in regressed_metric_revisions[pl_type]:
                    continue
                for revision, _, _, _ in revisions:
                    if revision in regressed_metric_revisions[pl_type][metric]:
                        print(
                            f"{revision} REGRESSION (INDEX: "
                            f"{regressed_metric_revisions[pl_type][metric][revision]})"
                        )
                    else:
                        print(f"{revision} NO-CHANGE")

        print("Overall result:")
        for revision, _, _, _ in revisions:
            if revision in all_regressing_revisions:
                print(f"{revision} REGRESSION")
            else:
                print(f"{revision} NO-CHANGE")

        return all_regressing_revisions, regressed_metric_revisions


if __name__ == "__main__":
    detector = RegressionDetector("/home/sparky/mozilla-source/detector-testing/")
    # detector.detect_changes(
    #     test_name="browsertime-tp6m-geckoview-sina-nofis",
    #     platform="test-android-hw-a51-11-0-aarch64-shippable-qr/opt",
    #     base_revision="b1dad9107c06df86f3107e58c27d88a50b6c53c1",
    #     new_revision="b1dad9107c06df86f3107e58c27d88a50b6c53c1",
    #     depth=20,
    #     skip_download=True,
    #     overwrite=False,
    # )
    detector.detect_changes(
        test_name="browsertime-tp6m-geckoview-sina-nofis",
        platform="test-android-hw-a51-11-0-aarch64-shippable-qr/opt",
        base_revision="9b8e7381f776fa84a55e0a764c2ed4dad82316c5",
        new_revision="b1dad9107c06df86f3107e58c27d88a50b6c53c1",
        # Depth of -1 means auto-computed (everything in between the two given revisions),
        # None is direct comparisons, anything else uses the new_revision as a start
        # and goes backwards from there.
        depth=-1,
        skip_download=True,
        overwrite=False,
    )



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import json
import pathlib
import requests

from sys import stdout
from time import sleep

try:
    from urllib.parse import urlencode
    from urllib.request import urlopen, urlretrieve
except ImportError:
    from urllib import urlencode, urlretrieve
    from urllib2 import urlopen


TASK_IDS = (
    "https://firefox-ci-tc.services.mozilla.com/api/index/v1/tasks/"
    + "gecko.v2.{}.revision.{}.taskgraph"
)

TASK_INFO = "https://firefox-ci-tc.services.mozilla.com/api/queue/v1/task/{}"

BRANCH_URLS = {"autoland": "https://hg.mozilla.org/integration/autoland"}

PUSHLOG_TMPL = "{}/json-pushes?version=2&startID={}&endID={}"

REVISION_JSON_URL = "{}/json-rev/{}"


def write_same_line(x, sleep_time=0.0001):
    stdout.write("\r%s" % str(x))
    stdout.flush()
    sleep(sleep_time)


def finish_same_line():
    stdout.write("\r  \r\n")


def get_json(url, params=None):
    if params is not None:
        url += "?" + urlencode(params)

    r = urlopen(url).read().decode("utf-8")

    return json.loads(r)


def _retrieve_from_cache(filename, cache):
    if cache is not None and pathlib.Path(cache, filename).exists():
        path = pathlib.Path(cache, filename)
        try:
            print(f"Found {path} in the cache")
            with path.open() as f:
                return json.load(f)
        except Exception:
            print(
                "Something is wrong with the cache file. Deleting, and "
                "downloading again..."
            )
            path.unlink()
    return None


def _store_to_cache(result, filename, cache):
    if cache is not None:
        path = pathlib.Path(cache, filename)
        print(f"Caching results as {path}")
        with path.open("w+") as f:
            json.dump(result, f)


def get_revision_json(revision, branch="autoland", cache=None):
    if branch not in BRANCH_URLS:
        raise Exception("Unkown project/branch given.")

    cached_filename = f"revision-json-{revision}-{branch}.json"
    cached_data = _retrieve_from_cache(cached_filename, cache)
    if cached_data is not None:
        return cached_data

    url = REVISION_JSON_URL.format(BRANCH_URLS.get(branch), revision)        
    print(f"Downloading {url}...")
    r = get_json(url)
    print(f"Finished downloading {url}")

    _store_to_cache(r, cached_filename, cache)
    return r


def get_pushes(project, end_id, depth, full_response=False, cache=None):
    """
    Modified version from here: 
    https://searchfox.org/mozilla-central/rev/
    4d2b1f753871ce514f9dccfc5b1b5e867f499229/taskcluster/
    gecko_taskgraph/actions/util.py#123-142
    """
    pushes = []
    print("Downloading pushes...")
    while True:
        start_id = max(end_id - depth, 0)
        pushlog_url = PUSHLOG_TMPL.format(BRANCH_URLS.get(project), start_id, end_id)
        cached_filename = f"pushlog-{project}-{start_id}-{end_id}.json"
        res = _retrieve_from_cache(cached_filename, cache)
        if res is None:
            print(f"Downloading {pushlog_url}...")
            r = requests.get(pushlog_url)
            r.raise_for_status()
            res = r.json()
            _store_to_cache(res, cached_filename, cache)

        pushes = pushes + list(res["pushes"].keys())
        if len(pushes) >= depth:
            break

        end_id = start_id - 1
        start_id -= depth
        if start_id < 0:
            break
        sleep(0.01)
    print("Finished downloading pushes")
    pushes = sorted(pushes)[-depth:]
    push_dict = {push: res["pushes"][push] for push in pushes}
    return push_dict if full_response else pushes


def find_task_group_id(revision, branch, search_crons=False, cache=None):
    # Find the task IDs from this revision first
    task_ids_url = TASK_IDS.format(branch, revision)

    cached_filename = f"task-ids-data-{branch}-{revision}.json"
    task_ids_data = _retrieve_from_cache(cached_filename, cache)
    if task_ids_data is None:
        print("Downloading task ids from: %s" % task_ids_url)
        task_ids_data = get_json(task_ids_url)
        _store_to_cache(task_ids_data, cached_filename, cache)
    if "tasks" not in task_ids_data or len(task_ids_data["tasks"]) == 0:
        raise Exception("Cannot find any task IDs for %s!" % revision)

    task_group_ids = []
    for task in task_ids_data["tasks"]:
        # Only find the task group ID for the decision task if we
        # don't need to search for cron tasks
        if not search_crons and not task["namespace"].endswith("decision"):
            continue
        task_group_url = TASK_INFO.format(task["taskId"])

        cached_filename = f"task-group-{task['taskId']}-{branch}-{revision}.json"
        task_info = _retrieve_from_cache(cached_filename, cache)
        if task_info is None:
            print("Downloading task group id from: %s" % task_group_url)
            task_info = get_json(task_group_url)
            _store_to_cache(task_info, cached_filename, cache)

        task_group_ids.append(task_info["taskGroupId"])

    return task_group_ids
